"""Static analysis subpackage for Gabion."""

from .dataflow_audit import (
    AnalysisResult,
    AuditConfig,
    analyze_paths,
    apply_baseline,
    build_refactor_plan,
    build_synthesis_plan,
    compute_violations,
    load_baseline,
    render_dot,
    render_refactor_plan,
    render_report,
    render_protocol_stubs,
    render_synthesis_section,
    resolve_baseline_path,
    write_baseline,
)

__all__ = [
    "AnalysisResult",
    "AuditConfig",
    "analyze_paths",
    "apply_baseline",
    "build_refactor_plan",
    "build_synthesis_plan",
    "compute_violations",
    "load_baseline",
    "render_dot",
    "render_refactor_plan",
    "render_report",
    "render_protocol_stubs",
    "render_synthesis_section",
    "resolve_baseline_path",
    "write_baseline",
]
